package org.apache.bcel.generic;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import org.apache.bcel.classfile.Constant;
import org.apache.bcel.classfile.ConstantCP;
import org.apache.bcel.classfile.ConstantClass;
import org.apache.bcel.classfile.ConstantDouble;
import org.apache.bcel.classfile.ConstantDynamic;
import org.apache.bcel.classfile.ConstantFieldref;
import org.apache.bcel.classfile.ConstantFloat;
import org.apache.bcel.classfile.ConstantInteger;
import org.apache.bcel.classfile.ConstantInterfaceMethodref;
import org.apache.bcel.classfile.ConstantInvokeDynamic;
import org.apache.bcel.classfile.ConstantLong;
import org.apache.bcel.classfile.ConstantMethodref;
import org.apache.bcel.classfile.ConstantNameAndType;
import org.apache.bcel.classfile.ConstantPool;
import org.apache.bcel.classfile.ConstantString;
import org.apache.bcel.classfile.ConstantUtf8;
import org.apache.bcel.classfile.Utility;

public class ConstantPoolGen {
   private static final int DEFAULT_BUFFER_SIZE = 256;
   private static final String METHODREF_DELIM = ":";
   private static final String IMETHODREF_DELIM = "#";
   private static final String FIELDREF_DELIM = "&";
   private static final String NAT_DELIM = "%";
   /** @deprecated */
   @Deprecated
   protected int size;
   /** @deprecated */
   @Deprecated
   protected Constant[] constants;
   /** @deprecated */
   @Deprecated
   protected int index;
   private final Map<String, Integer> stringTable;
   private final Map<String, Integer> classTable;
   private final Map<String, Integer> utf8Table;
   private final Map<String, Integer> natTable;
   private final Map<String, Integer> cpTable;

   public ConstantPoolGen() {
      this.index = 1;
      this.stringTable = new HashMap();
      this.classTable = new HashMap();
      this.utf8Table = new HashMap();
      this.natTable = new HashMap();
      this.cpTable = new HashMap();
      this.size = 256;
      this.constants = new Constant[this.size];
   }

   public ConstantPoolGen(Constant[] cs) {
      this.index = 1;
      this.stringTable = new HashMap();
      this.classTable = new HashMap();
      this.utf8Table = new HashMap();
      this.natTable = new HashMap();
      this.cpTable = new HashMap();
      StringBuilder sb = new StringBuilder(256);
      this.size = Math.min(Math.max(256, cs.length + 64), 65536);
      this.constants = new Constant[this.size];
      System.arraycopy(cs, 0, this.constants, 0, cs.length);
      if (cs.length > 0) {
         this.index = cs.length;
      }

      for(int i = 1; i < this.index; ++i) {
         Constant c = this.constants[i];
         ConstantUtf8 u8;
         String key;
         if (c instanceof ConstantString) {
            ConstantString s = (ConstantString)c;
            u8 = (ConstantUtf8)this.constants[s.getStringIndex()];
            key = u8.getBytes();
            if (!this.stringTable.containsKey(key)) {
               this.stringTable.put(key, i);
            }
         } else if (c instanceof ConstantClass) {
            ConstantClass s = (ConstantClass)c;
            u8 = (ConstantUtf8)this.constants[s.getNameIndex()];
            key = u8.getBytes();
            if (!this.classTable.containsKey(key)) {
               this.classTable.put(key, i);
            }
         } else {
            ConstantUtf8 u8;
            if (c instanceof ConstantNameAndType) {
               ConstantNameAndType n = (ConstantNameAndType)c;
               u8 = (ConstantUtf8)this.constants[n.getNameIndex()];
               u8 = (ConstantUtf8)this.constants[n.getSignatureIndex()];
               sb.append(u8.getBytes());
               sb.append("%");
               sb.append(u8.getBytes());
               String key = sb.toString();
               sb.delete(0, sb.length());
               if (!this.natTable.containsKey(key)) {
                  this.natTable.put(key, i);
               }
            } else {
               String className;
               if (c instanceof ConstantUtf8) {
                  ConstantUtf8 u = (ConstantUtf8)c;
                  className = u.getBytes();
                  if (!this.utf8Table.containsKey(className)) {
                     this.utf8Table.put(className, i);
                  }
               } else if (c instanceof ConstantCP) {
                  ConstantCP m = (ConstantCP)c;
                  if (c instanceof ConstantInvokeDynamic) {
                     className = Integer.toString(((ConstantInvokeDynamic)m).getBootstrapMethodAttrIndex());
                  } else if (c instanceof ConstantDynamic) {
                     className = Integer.toString(((ConstantDynamic)m).getBootstrapMethodAttrIndex());
                  } else {
                     ConstantClass clazz = (ConstantClass)this.constants[m.getClassIndex()];
                     u8 = (ConstantUtf8)this.constants[clazz.getNameIndex()];
                     className = Utility.pathToPackage(u8.getBytes());
                  }

                  ConstantNameAndType n = (ConstantNameAndType)this.constants[m.getNameAndTypeIndex()];
                  u8 = (ConstantUtf8)this.constants[n.getNameIndex()];
                  String methodName = u8.getBytes();
                  u8 = (ConstantUtf8)this.constants[n.getSignatureIndex()];
                  String signature = u8.getBytes();
                  String delim = ":";
                  if (c instanceof ConstantInterfaceMethodref) {
                     delim = "#";
                  } else if (c instanceof ConstantFieldref) {
                     delim = "&";
                  }

                  sb.append(className);
                  sb.append(delim);
                  sb.append(methodName);
                  sb.append(delim);
                  sb.append(signature);
                  String key = sb.toString();
                  sb.delete(0, sb.length());
                  if (!this.cpTable.containsKey(key)) {
                     this.cpTable.put(key, i);
                  }
               }
            }
         }
      }

   }

   public ConstantPoolGen(ConstantPool cp) {
      this(cp.getConstantPool());
   }

   public int addArrayClass(ArrayType type) {
      return this.addClass_(type.getSignature());
   }

   public int addClass(ObjectType type) {
      return this.addClass(type.getClassName());
   }

   public int addClass(String str) {
      return this.addClass_(str.replace('.', '/'));
   }

   private int addClass_(String clazz) {
      int cpRet;
      if ((cpRet = this.lookupClass(clazz)) != -1) {
         return cpRet;
      } else {
         this.adjustSize();
         ConstantClass c = new ConstantClass(this.addUtf8(clazz));
         int ret = this.index;
         this.constants[this.index++] = c;
         return this.computeIfAbsent(this.classTable, clazz, ret);
      }
   }

   public int addConstant(Constant c, ConstantPoolGen cp) {
      Constant[] constants = cp.getConstantPool().getConstantPool();
      ConstantUtf8 u8;
      switch (c.getTag()) {
         case 1:
            return this.addUtf8(((ConstantUtf8)c).getBytes());
         case 2:
         default:
            throw new IllegalArgumentException("Unknown constant type " + c);
         case 3:
            return this.addInteger(((ConstantInteger)c).getBytes());
         case 4:
            return this.addFloat(((ConstantFloat)c).getBytes());
         case 5:
            return this.addLong(((ConstantLong)c).getBytes());
         case 6:
            return this.addDouble(((ConstantDouble)c).getBytes());
         case 7:
            ConstantClass s = (ConstantClass)c;
            u8 = (ConstantUtf8)constants[s.getNameIndex()];
            return this.addClass(u8.getBytes());
         case 8:
            ConstantString s = (ConstantString)c;
            u8 = (ConstantUtf8)constants[s.getStringIndex()];
            return this.addString(u8.getBytes());
         case 9:
         case 10:
         case 11:
            ConstantCP m = (ConstantCP)c;
            ConstantClass clazz = (ConstantClass)constants[m.getClassIndex()];
            ConstantNameAndType n = (ConstantNameAndType)constants[m.getNameAndTypeIndex()];
            ConstantUtf8 u8 = (ConstantUtf8)constants[clazz.getNameIndex()];
            String className = Utility.pathToPackage(u8.getBytes());
            u8 = (ConstantUtf8)constants[n.getNameIndex()];
            String name = u8.getBytes();
            u8 = (ConstantUtf8)constants[n.getSignatureIndex()];
            String signature = u8.getBytes();
            switch (c.getTag()) {
               case 9:
                  return this.addFieldref(className, name, signature);
               case 10:
                  return this.addMethodref(className, name, signature);
               case 11:
                  return this.addInterfaceMethodref(className, name, signature);
               default:
                  throw new IllegalArgumentException("Unknown constant type " + c);
            }
         case 12:
            ConstantNameAndType n = (ConstantNameAndType)c;
            u8 = (ConstantUtf8)constants[n.getNameIndex()];
            ConstantUtf8 u8_2 = (ConstantUtf8)constants[n.getSignatureIndex()];
            return this.addNameAndType(u8.getBytes(), u8_2.getBytes());
      }
   }

   public int addDouble(double n) {
      int ret;
      if ((ret = this.lookupDouble(n)) != -1) {
         return ret;
      } else {
         this.adjustSize();
         ret = this.index;
         this.constants[this.index] = new ConstantDouble(n);
         this.index += 2;
         return ret;
      }
   }

   public int addFieldref(String className, String fieldName, String signature) {
      int cpRet;
      if ((cpRet = this.lookupFieldref(className, fieldName, signature)) != -1) {
         return cpRet;
      } else {
         this.adjustSize();
         int classIndex = this.addClass(className);
         int nameAndTypeIndex = this.addNameAndType(fieldName, signature);
         int ret = this.index;
         this.constants[this.index++] = new ConstantFieldref(classIndex, nameAndTypeIndex);
         return this.computeIfAbsent(this.cpTable, className + "&" + fieldName + "&" + signature, ret);
      }
   }

   public int addFloat(float n) {
      int ret;
      if ((ret = this.lookupFloat(n)) != -1) {
         return ret;
      } else {
         this.adjustSize();
         ret = this.index;
         this.constants[this.index++] = new ConstantFloat(n);
         return ret;
      }
   }

   public int addInteger(int n) {
      int ret;
      if ((ret = this.lookupInteger(n)) != -1) {
         return ret;
      } else {
         this.adjustSize();
         ret = this.index;
         this.constants[this.index++] = new ConstantInteger(n);
         return ret;
      }
   }

   public int addInterfaceMethodref(MethodGen method) {
      return this.addInterfaceMethodref(method.getClassName(), method.getName(), method.getSignature());
   }

   public int addInterfaceMethodref(String className, String methodName, String signature) {
      int cpRet;
      if ((cpRet = this.lookupInterfaceMethodref(className, methodName, signature)) != -1) {
         return cpRet;
      } else {
         this.adjustSize();
         int classIndex = this.addClass(className);
         int nameAndTypeIndex = this.addNameAndType(methodName, signature);
         int ret = this.index;
         this.constants[this.index++] = new ConstantInterfaceMethodref(classIndex, nameAndTypeIndex);
         return this.computeIfAbsent(this.cpTable, className + "#" + methodName + "#" + signature, ret);
      }
   }

   public int addLong(long n) {
      int ret;
      if ((ret = this.lookupLong(n)) != -1) {
         return ret;
      } else {
         this.adjustSize();
         ret = this.index;
         this.constants[this.index] = new ConstantLong(n);
         this.index += 2;
         return ret;
      }
   }

   public int addMethodref(MethodGen method) {
      return this.addMethodref(method.getClassName(), method.getName(), method.getSignature());
   }

   public int addMethodref(String className, String methodName, String signature) {
      int cpRet;
      if ((cpRet = this.lookupMethodref(className, methodName, signature)) != -1) {
         return cpRet;
      } else {
         this.adjustSize();
         int nameAndTypeIndex = this.addNameAndType(methodName, signature);
         int classIndex = this.addClass(className);
         int ret = this.index;
         this.constants[this.index++] = new ConstantMethodref(classIndex, nameAndTypeIndex);
         return this.computeIfAbsent(this.cpTable, className + ":" + methodName + ":" + signature, ret);
      }
   }

   public int addNameAndType(String name, String signature) {
      int ret;
      if ((ret = this.lookupNameAndType(name, signature)) != -1) {
         return ret;
      } else {
         this.adjustSize();
         int nameIndex = this.addUtf8(name);
         int signatureIndex = this.addUtf8(signature);
         ret = this.index;
         this.constants[this.index++] = new ConstantNameAndType(nameIndex, signatureIndex);
         return this.computeIfAbsent(this.natTable, name + "%" + signature, ret);
      }
   }

   public int addString(String str) {
      int ret;
      if ((ret = this.lookupString(str)) != -1) {
         return ret;
      } else {
         int utf8 = this.addUtf8(str);
         this.adjustSize();
         ConstantString s = new ConstantString(utf8);
         ret = this.index;
         this.constants[this.index++] = s;
         return this.computeIfAbsent(this.stringTable, str, ret);
      }
   }

   public int addUtf8(String n) {
      int ret;
      if ((ret = this.lookupUtf8(n)) != -1) {
         return ret;
      } else {
         this.adjustSize();
         ret = this.index;
         this.constants[this.index++] = new ConstantUtf8(n);
         return this.computeIfAbsent(this.utf8Table, n, ret);
      }
   }

   protected void adjustSize() {
      if (this.index + 3 >= 65536) {
         throw new IllegalStateException("The number of constants " + (this.index + 3) + " is over the size of the constant pool: " + '\uffff');
      } else {
         if (this.index + 3 >= this.size) {
            Constant[] cs = this.constants;
            this.size *= 2;
            this.size = Math.min(this.size, 65536);
            this.constants = new Constant[this.size];
            System.arraycopy(cs, 0, this.constants, 0, this.index);
         }

      }
   }

   private int computeIfAbsent(Map<String, Integer> map, String key, int value) {
      return (Integer)map.computeIfAbsent(key, (k) -> {
         return value;
      });
   }

   public Constant getConstant(int i) {
      return this.constants[i];
   }

   public ConstantPool getConstantPool() {
      return new ConstantPool(this.constants);
   }

   public ConstantPool getFinalConstantPool() {
      return new ConstantPool((Constant[])Arrays.copyOf(this.constants, this.index));
   }

   private int getIndex(Map<String, Integer> map, String key) {
      return this.toIndex((Integer)map.get(key));
   }

   public int getSize() {
      return this.index;
   }

   public int lookupClass(String str) {
      return this.getIndex(this.classTable, str.replace('.', '/'));
   }

   public int lookupDouble(double n) {
      long bits = Double.doubleToLongBits(n);

      for(int i = 1; i < this.index; ++i) {
         if (this.constants[i] instanceof ConstantDouble) {
            ConstantDouble c = (ConstantDouble)this.constants[i];
            if (Double.doubleToLongBits(c.getBytes()) == bits) {
               return i;
            }
         }
      }

      return -1;
   }

   public int lookupFieldref(String className, String fieldName, String signature) {
      return this.getIndex(this.cpTable, className + "&" + fieldName + "&" + signature);
   }

   public int lookupFloat(float n) {
      int bits = Float.floatToIntBits(n);

      for(int i = 1; i < this.index; ++i) {
         if (this.constants[i] instanceof ConstantFloat) {
            ConstantFloat c = (ConstantFloat)this.constants[i];
            if (Float.floatToIntBits(c.getBytes()) == bits) {
               return i;
            }
         }
      }

      return -1;
   }

   public int lookupInteger(int n) {
      for(int i = 1; i < this.index; ++i) {
         if (this.constants[i] instanceof ConstantInteger) {
            ConstantInteger c = (ConstantInteger)this.constants[i];
            if (c.getBytes() == n) {
               return i;
            }
         }
      }

      return -1;
   }

   public int lookupInterfaceMethodref(MethodGen method) {
      return this.lookupInterfaceMethodref(method.getClassName(), method.getName(), method.getSignature());
   }

   public int lookupInterfaceMethodref(String className, String methodName, String signature) {
      return this.getIndex(this.cpTable, className + "#" + methodName + "#" + signature);
   }

   public int lookupLong(long n) {
      for(int i = 1; i < this.index; ++i) {
         if (this.constants[i] instanceof ConstantLong) {
            ConstantLong c = (ConstantLong)this.constants[i];
            if (c.getBytes() == n) {
               return i;
            }
         }
      }

      return -1;
   }

   public int lookupMethodref(MethodGen method) {
      return this.lookupMethodref(method.getClassName(), method.getName(), method.getSignature());
   }

   public int lookupMethodref(String className, String methodName, String signature) {
      return this.getIndex(this.cpTable, className + ":" + methodName + ":" + signature);
   }

   public int lookupNameAndType(String name, String signature) {
      return this.getIndex(this.natTable, name + "%" + signature);
   }

   public int lookupString(String str) {
      return this.getIndex(this.stringTable, str);
   }

   public int lookupUtf8(String n) {
      return this.getIndex(this.utf8Table, n);
   }

   public void setConstant(int i, Constant c) {
      this.constants[i] = c;
   }

   private int toIndex(Integer index) {
      return index != null ? index : -1;
   }

   public String toString() {
      StringBuilder buf = new StringBuilder();

      for(int i = 1; i < this.index; ++i) {
         buf.append(i).append(")").append(this.constants[i]).append("\n");
      }

      return buf.toString();
   }
}
